--[[
	Free Layout for Grid2 ( Created by Michael)
	
	*Enable a especial layout called "Free Layout", in free layout mode
	 Grid2 unit frames can be rearranged using drag and drop, to move a unit frame
	 simply push a modifier key combination + Left Click over a unit frame.
	*Provides a dropdown menu to change layout and to configure "Free Layout" layout, 
	 menu is activated by a modifier key combination + Right Click over the Grid2 frames
 	*Modifiers key combination must be configured in Grid2Options ("Misc" section)
--]]

local Grid2Layout= Grid2Layout
local Grid2Frame= Grid2Frame

local fmt= string.format

local SettingUnitsPerColumn = 5
local SettingMaxColumns = 5
local SettingSortType = "group"
local SettingSortTanks = false
local SettingMaxGroup = nil

local curLayoutName
local nameListStr
local nameList,newList= {}, {}
local maxGroup
local frameLayout 
local frameUnit 
local dragUnitName
local queueUpdate
local CheckModifierFunc
local Events= {}

---
--- "Free Layout" layout management
---

local FreeLayout= {
	meta = {
		raid25 = true,
		raid20 = true,
		raid15 = true,
		raid10 = true,
		party = true,
		solo = true,
	},
	defaults = {
        allowVehicleTarget = true,
		toggleForVehicle = true,
		showPlayer = true,
		showParty = true,
		showRaid = true,
		showSolo = true,
	},
	[1] = {
		nameList = "",
		unitsPerColumn = 5,
		maxColumns = 5,
		sortMethod = "NAMELIST"
	},
}
Grid2Layout:AddLayout("Free Layout",FreeLayout)


local function LayoutUpdate(NoResize)
	if InCombatLockdown() then
		queueUpdate= 1
		return
	end	
	local header= Grid2Layout.groups.raid[1] or Grid2Layout.groups.party[1]
	if header then 	
		local visible= header:IsVisible()
		if visible then	header:Hide() end
		header:SetAttribute("nameList", nameListStr )
		header:SetAttribute("maxColumns", SettingMaxColumns )
		header:SetAttribute("unitsPerColumn", SettingUnitsPerColumn )
		if visible then header:Show() end
		if not NoResize then
			Grid2Layout:UpdateSize()
		end
	end	
	queueUpdate= false
	print("Layout Update")
end

local UpdateRoster
do
	local PartyTypeUnits= {raid25= 25, raid20= 20, raid15=15, raid10=10, party=5, solo=1}
	local TankKey= { ["MAINTANK"]= 0, ["MAINASSIST"]=1 }
	local SortKeyGen= {
		["group"]     = function(n,g)     return fmt("%d%s",g,n) end,
		["grouptank"] = function(n,g,_,r) return fmt("%d%d%s", g, TankKey[r] or 1,n) end,
		["class"]     = function(_,_,c  ) return c end,
		["classtank"] = function(n,_,c,r) return fmt("%d%s%s", TankKey[r] or 1, c, n) end,
		["name" ]     = function(n)       return n end,
		["nametank"]  = function(n,_,_,r) return fmt("%d%s", TankKey[r] or 1, n) end,
		["user"]      = function(n,_,_,r) return nameList[n] or "z" end,
		["usertank"]  = function(n,_,_,r) return nameList[n] or "z" end,
	}
	local function GetPartyRosterInfo(i)
		local unit= i<2 and "player" or "party"..(i-1)
		local name= UnitName(unit)
		if name and name~="" then
			local _,class= UnitClass(unit)
			local role= GetPartyAssignment("MAINTANK", unit) and "MAINTANK" or ""
			return name,nil,1,nil,nil,class,nil,nil,nil,role
		end
	end
	local function GetNumGroupMembers()
		local count= GetNumRaidMembers()
		if count>0 then
			return count, GetRaidRosterInfo
		else
			return GetNumPartyMembers()+1, GetPartyRosterInfo
		end
	end
	UpdateRoster= function(updateLayout)
		local SortKey= SortKeyGen[SettingSortTanks and SettingSortType.."tank" or SettingSortType]
		local maxCount= PartyTypeUnits[Grid2Layout.partyType] or 5 
		maxGroup= SettingMaxGroup or math.ceil( maxCount / 5 )
		
		local count,GetGroupRosterInfo = GetNumGroupMembers()
		for i=1,count do
			local name,_,group,_,_,class,_,_,_,role= GetGroupRosterInfo(i)
			if name then 
				if group<=maxGroup then
					newList[#newList+1]= name
					newList[name]= SortKey(name,group,class,role)
				end
			else
				break
			end
		end
		
		table.sort( newList, function(a,b) return newList[a] < newList[b] end )
		
		local modified
		
		local newListStr= table.concat( newList, "," )
		if newListStr ~= nameListStr then
			nameList,newList,nameListStr= newList,nameList,newListStr
			for i=1,#nameList do
				nameList[ nameList[i] ]= i
			end
			modified= true
		end
		wipe(newList)

		local maxColumns= math.ceil( maxCount / SettingUnitsPerColumn )
		if maxColumns ~= SettingMaxColumns then
			SettingMaxColumns= maxColumns
			modified= true
		end

		if updateLayout and modified then
			if updateLayout==0 then
				LayoutUpdate()
			else
				queueUpdate= updateLayout
			end
		end
		print("Update Roster")
	end
end

local function LayoutSwapUnits(n1,n2)
	local i1,i2= nameList[n1],nameList[n2]
	if i1 and i2 then
		nameList[i1], nameList[i2]= n2, n1
		nameList[n1], nameList[n2]= i2, i1
	end
	nameListStr= table.concat(nameList,",")
	SettingSortType= "user"
	LayoutUpdate(true)
end

local KeyModifiers= { 
	SHIFT= "IsShiftKeyDown()", LSHIFT= "IsLeftShiftKeyDown()",RSHIFT= "IsRightShiftKeyDown()", 
	ALT= "IsAltKeyDown()",LALT= "IsLeftAltKeyDown()", RALT= "IsRightAltKeyDown()", 
	CTRL= "IsControlKeyDown()",LCTRL= "IsLeftControlKeyDown()", RCTRL= "IsRightControlKeyDown()"
}
local function LayoutSetModifier(settings) 
	local s={}
	for k,v in pairs(KeyModifiers) do
		if settings[k] then
			s[#s+1]= v
		end
	end
	if (#s==0) then s[1]= "false" end
	s= "return function() return " .. table.concat(s, " and ") .. " end"
	CheckModifierFunc = assert(loadstring(s))()
end

function Events:RAID_ROSTER_UPDATE()
	print("Raid roster update")
	UpdateRoster(2)
end

function Events:PARTY_MEMBERS_CHANGED()
	print("Party members changed")
	SettingMaxGroup= nil -- Reset user defined visible groups
	UpdateRoster(2)
end

local function LayoutPrepareForLoad()
    UpdateRoster()
	FreeLayout[1].nameList= nameListStr
	FreeLayout[1].maxColumns= SettingMaxColumns
	FreeLayout[1].unitsPerColumn= SettingUnitsPerColumn
	frameLayout:RegisterEvent("PARTY_MEMBERS_CHANGED")
	frameLayout:RegisterEvent("RAID_ROSTER_UPDATE")
end

local function LayoutPrepareForUnload()
	frameLayout:UnregisterEvent("RAID_ROSTER_UPDATE")
	frameLayout:UnregisterEvent("PARTY_MEMBERS_CHANGED")
end

---
--- Popup Menu management
---

local LayoutMenuShow
do
	local partyType
	local menuFrame
	local menuFree
	local menuLayouts = {}

	local function MenuGetLayouts(partyType)
		local list={}
		for name, layout in pairs(Grid2Layout.layoutSettings) do
		  if layout.meta[partyType] then
			 list[#list+1]= name
		  end
		end
		table.sort(list)
		return list
	end

	local function MenuSetLayout(self)
		if self.value ~= curLayoutName then
			if curLayoutName=="Free Layout" then
				ToggleDropDownMenu(1, nil, menuFrame, "cursor",0,0, menuFree);
			end
			Grid2Layout:LoadLayout(self.value)
		end	
	end
	
	local function MenuSetUnits(self)
		SettingUnitsPerColumn= tonumber(self.value)
		LayoutUpdate() 
	end
	
	local function MenuSetSort(self)
		SettingSortType= self.value
		UpdateRoster(0)
	end

	local function MenuSetGroup(self)
		SettingMaxGroup= tonumber(self.value)
		UpdateRoster(0)
	end

	local function MenuSetTank() 
		SettingSortTanks = not SettingSortTanks 
		UpdateRoster(0) 
	end 

	local function MenuCreateLayouts()
		if partyType ~= Grid2Layout.partyType then
			partyType= Grid2Layout.partyType
			wipe(menuLayouts)
			menuLayouts[1]= { text = "Select Layout",  notCheckable= true, isTitle = true}
			local list= MenuGetLayouts(partyType)
			for _,name in ipairs(list) do
				if name ~= "None" then
					local option= { text = name, checked= function() return curLayoutName==name end, func = MenuSetLayout }
					table.insert(menuLayouts, name=="Free Layout" and 2 or #menuLayouts+1,option)
				end
			end
		end	
	end
	
	local function MenuCreateFree()
		if not menuFree then
			local unitMenu= {}
			for i=5,25,5 do
				unitMenu[#unitMenu+1]= { text = tostring(i), checked= function() return SettingUnitsPerColumn==i end, func = MenuSetUnits }
			end
			local groupMenu= {}
			for i=1,8 do
				groupMenu[#groupMenu+1]= { text = tostring(i), checked= function() return maxGroup>=i end, func = MenuSetGroup }
			end
			menuFree= {
				{ text = "Free Layout Menu",  notCheckable= true, isTitle = true},
				{ text = "Sort by Group", value="group", checked= function() return SettingSortType=="group" end, func = MenuSetSort },
				{ text = "Sort by Class", value="class", checked= function() return SettingSortType=="class" end, func = MenuSetSort },
				{ text = "Sort by Name",  value="name",  checked= function() return SettingSortType=="name"  end, func = MenuSetSort },
				{ text = "Tanks First",   				 checked= function() return SettingSortTanks         end, func = MenuSetTank },
				{ text = "Units per column", notCheckable= true, hasArrow = true, menuList = unitMenu },
				{ text = "Visible groups", notCheckable= true, hasArrow = true, menuList = groupMenu },			
				{ text = "Change Layout", notCheckable= true, hasArrow = true, menuList = menuLayouts },			
			}
		end
	end

	LayoutMenuShow= function()
		menuFrame= menuFrame or CreateFrame("Frame", "Grid2FreeLayoutMenu", UIParent, "UIDropDownMenuTemplate") 
		MenuCreateLayouts()
		MenuCreateFree()
		EasyMenu((curLayoutName=="Free Layout" and menuFree or menuLayouts), menuFrame, "cursor", 0 , 0, "MENU", 0)
	end
end

---
--- Unit frames mouse drag management
---

local function LayoutDragGetCursorPosition()
	local s = frameLayout:GetEffectiveScale()
	local x, y = GetCursorPosition()
	return x/s,y/s
end

local function LayoutDragGetUnitFromPoint(x,y)
	for _, frame in next, Grid2Frame.registeredFrames do
		if frame:IsVisible() and x>=frame:GetLeft() and x<=frame:GetRight() and y>=frame:GetBottom() and y<=frame:GetTop() then
			return frame.unit and UnitName(frame.unit) or nil, frame
		end		
	end
end

local function LayoutDragStart(frame)
	frameUnit:SetPoint("TOPLEFT",frame, "TOPLEFT")
	frameUnit:Show()
	frameUnit:StartMoving() 
end

local function LayoutDragStop()
	dragUnitName= nil	
    frameUnit:StopMovingOrSizing()
	frameUnit:Hide()
	frameUnit:ClearAllPoints()
end

local function LayoutDragOnMouseDown(button)
    if button=="LeftButton" then
		local frame
		local x,y= LayoutDragGetCursorPosition()
		dragUnitName,frame= LayoutDragGetUnitFromPoint(x,y)
		if dragUnitName then
			if not frameUnit then
				frameUnit= CreateFrame("Frame", nil, frameLayout)
				frameUnit:SetMovable(true)
				frameUnit:SetBackdrop({
						edgeFile = "Interface\\Addons\\Grid2\\white16x16", edgeSize = 2,
						insets = {left = 2, right = 2, top = 2, bottom = 2},
				})
				frameUnit:SetBackdropBorderColor(1,1,1,1)
				frameUnit:SetBackdropColor(0,0,0,0)
			end
			frameUnit:SetSize( frame:GetSize() )
			LayoutDragStart(frame)
		end	
	end	
end

local function LayoutDragOnMouseUp(button)
	local x= (frameUnit:GetLeft() + frameUnit:GetRight()) / 2
	local y= (frameUnit:GetTop() + frameUnit:GetBottom()) / 2
	local unitName= LayoutDragGetUnitFromPoint(x,y)
	if unitName and (dragUnitName ~= unitName) then
		LayoutSwapUnits(unitName,dragUnitName)
	end
	print("DRAGSTOP:", unitName, " - ", nameListStr)
	LayoutDragStop()
end

---
--- Layout manager
---

local function LayoutManagerOnMouseDownEvent(_, button)
	LayoutDragOnMouseDown(button)
end

local function LayoutManagerOnMouseUpEvent(_, button)
	if dragUnitName then
		LayoutDragOnMouseUp(button)
	elseif button=="RightButton" then
		LayoutMenuShow()
	end
end

local function LayoutManagerShow()
	local Grid2LayoutFrame= Grid2LayoutFrame
	frameLayout:SetFrameLevel( Grid2LayoutFrame:GetFrameLevel()+25 )
	frameLayout:SetScale(Grid2LayoutFrame:GetScale())
	frameLayout:EnableMouse(true)

	local border= Grid2Layout.db.profile.Spacing*2
	local width,height= Grid2LayoutFrame:GetSize()
	frameLayout:SetSize( width-border, height-border )

	if curLayoutName=="Free Layout" then
		frameLayout:SetScript("OnMouseDown", LayoutManagerOnMouseDownEvent)
	end
end

local function LayoutManagerHide()
	if dragUnitName then LayoutDragStop() end
	frameLayout:SetScript("OnMouseDown", nil)
	frameLayout:EnableMouse(false)
	frameLayout:SetSize(10,10)
end

local LayoutManagerOnUpdateEvent
do
	local oldmodifier
	local timerElapsed= 0.25
	local timerFrequency= 0.25
	LayoutManagerOnUpdateEvent= function(self, elapsed)
		timerElapsed= timerElapsed - elapsed
		if timerElapsed<0 then
			local modifier= CheckModifierFunc()
			if modifier ~= oldmodifier then
				oldmodifier= modifier
				if modifier then
					LayoutManagerShow()
				else
					LayoutManagerHide()
				end
			end
			if queueUpdate then
				queueUpdate= queueUpdate-1
				if queueUpdate<=0 then
					LayoutUpdate()
				end		
			end
			timerElapsed= timerFrequency
		end
	end
end

function LayoutManagerEnable(CombatLock)
	frameLayout:SetScript("OnUpdate", LayoutManagerOnUpdateEvent)
	frameLayout:SetScript("OnMouseUp", LayoutManagerOnMouseUpEvent)
	if not CombatLock then
		frameLayout:SetScript("OnEvent", function(self,event,...) Events[event](self,...) end)	
		frameLayout:RegisterEvent("PLAYER_REGEN DISABLED")
		frameLayout:RegisterEvent("PLAYER_REGEN_ENABLED")	
	end
	frameLayout:Show()
end	

function LayoutManagerDisable(CombatLock)
	LayoutManagerHide() 
	frameLayout:Hide()
	frameLayout:SetScript("OnUpdate", nil)
	frameLayout:SetScript("OnMouseUp", nil)
	frameLayout:SetScript("OnMouseDown", nil)
	if not CombatLock then
		frameLayout:SetScript("OnEvent",nil)
		frameLayout:UnregisterEvent("PLAYER_REGEN DISABLED")
		frameLayout:UnregisterEvent("PLAYER_REGEN_ENABLED")	
	end
end
	
function Events:PLAYER_REGEN_ENABLED()
	LayoutManagerEnable(true)
	print("REGEN ENABLED")
end
		
function Events:PLAYER_REGEN_DISABLED()
	LayoutManagerDisable(true)
	print("REGEN DISABLED")
end

---
---
---

function LayoutInit()
	local Grid2LayoutFrame= Grid2LayoutFrame
	if not frameLayout then
		frameLayout= CreateFrame("Frame", nil, Grid2LayoutFrame)
		frameLayout:SetPoint("CENTER", Grid2LayoutFrame, "CENTER")
	end
	local dbx= Grid2Layout.db.profile
	if not dbx.FreeLayoutKeys then dbx.FreeLayoutKeys= {} end
	LayoutSetModifier( dbx.FreeLayoutKeys )
	LayoutManagerEnable()
end	

---
--- Methods to configure Free Layout from options
---	

function Grid2Layout:FreeLayoutSetKeys(keys)
	LayoutSetModifier( keys  ) 
end	
	
---
--- Hook to activate Free Layout
---

local prev_LoadLayout= Grid2Layout.LoadLayout
function Grid2Layout:LoadLayout(layoutName)
	if not frameLayout then LayoutInit() end
	if curLayoutName~=layoutName then
		if layoutName=="Free Layout" then
			LayoutPrepareForLoad()
		elseif curLayoutName== "Free Layout" then
			LayoutPrepareForUnload()
		end
	end	
	prev_LoadLayout(self,layoutName)
	curLayoutName= layoutName
end


---
--- Gri2Options configuration
---

local prev_LoadOptions = Grid2.LoadOptions
function Grid2:LoadOptions()
	prev_LoadOptions()
	local FreeKeys= Grid2Layout.db.profile.FreeLayoutKeys
	local options= {}
	for key,_ in pairs(KeyModifiers) do
		options[key]= {
			type = "toggle",
			name = key,
			width= "half",
			get = function() return FreeKeys[key] end,
			set = function(_,v) 
					FreeKeys[key] = v and true or nil 
					Grid2Layout:FreeLayoutSetKeys( FreeKeys ) 
			end,
		}
	end
	Grid2Options:AddModuleOptions("Misc","Free Layout: select activation keys", options)
end
